﻿using Microsoft.Data.OData.Query;
using Microsoft.Data.OData.Query.SemanticAst;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Net;
using System.Net.Http;
using System.Reflection;
using System.Text.RegularExpressions;
using System.Web.Http;
using System.Web.Http.OData;
using System.Web.Http.OData.Query;

namespace test_mvc_webapi_ef_ng.Controllers
{
    [RoutePrefix("api/app")]
    public class AppController : ApiController
    {
        private Models.Database _db = null;

        public AppController(Models.Database db)
        {
            _db = db;
        }

        // GET api/app
        [HttpGet]
        [Route("")]
        public PageResult Get(ODataQueryOptions<db.SystemInstance> parameters)
        {
            var data = _db.QueryAll();

            if (parameters.Filter != null && parameters.InlineCount == null)
            {
                var filtered = parameters.Filter.ApplyTo(data.QueryResult, new ODataQuerySettings());

                return select<db.SystemInstance, string>(parameters.SelectExpand.RawSelect, filtered.Cast<db.SystemInstance>());
            }

            if (parameters.Filter == null && parameters.InlineCount == null)
            {
                var selected = parameters.SelectExpand.ApplyTo(data.QueryResult.Select(d => d.SomeComplex.Name), new ODataQuerySettings());
            }

            var query = parameters.ApplyTo(data.QueryResult);

            var count = parameters.Filter != null ? parameters.Request.GetInlineCount() : parameters.InlineCount.GetEntityCount(data.QueryResult);

            return createPageResult<db.SystemInstance>(query, count);
        }

        private static PageResult<TResult> select<TQuery, TResult>(string selectedProperty, IQueryable<TQuery> filtered)
        {
            var selectionParameter = Expression.Parameter(typeof(TQuery));

            var properties = selectedProperty.Split('/');

            Expression selectionBody = null;
            foreach (var property in properties)
            {
                selectionBody = Expression.Property(selectionBody ?? selectionParameter, property);
            }

            var selectionProperty = Expression.Lambda<Func<TQuery, TResult>>(selectionBody, selectionParameter);

            var selected = filtered.Select(selectionProperty);
            var count = selected.Count();

            return new PageResult<TResult>(selected, null, count);
        }

        private static PageResult<TResult> createPageResult<TResult>(IQueryable query, long? count)
        {
            var method = typeof(AppController).GetMethod("castQuery", BindingFlags.Static | BindingFlags.NonPublic);
            method = method.MakeGenericMethod(query.ElementType, typeof(TResult));
            return method.Invoke(null, new object[] { query, count }) as PageResult<TResult>;
        }

        private static PageResult<TResult> castQuery<TQuery, TResult>(IQueryable<TQuery> query, long? count)
        {
            var cast = new PageResult<TQuery>(query as IEnumerable<TQuery>, null, count);

            var result = new List<TResult>();
            var e = cast.Items.GetEnumerator();
            while (e.MoveNext())
            {
                var c = e.Current;
                var item = (TResult)c.GetType().GetProperty("Instance").GetValue(c);
                result.Add(item);
            }

            return new PageResult<TResult>(result, null, count);
        }

        // GET api/app/some
        [HttpGet]
        [Route("some")]
        public PageResult GetSome(ODataQueryOptions<db.SomeComplex> parameters)
        {
            var data = _db.QuerySome();

            //var query = parameters.ApplyTo(data.QueryResult);

            return new PageResult<db.SomeComplex>(data.QueryResult, null, data.Count);
        }

        // POST api/app
        [HttpPost]
        [Route("")]
        public void Post(db.SystemInstance system)
        {
            _db.Insert(system);
        }

        // PUT api/app
        [HttpPut]
        [Route("")]
        public void Put(db.SystemInstance system)
        {
            _db.Update(system);
        }

        // DELETE api/app/1
        [HttpDelete]
        [Route("{id}")]
        public void Delete(int id)
        {
            _db.Delete(id);
        }
    }


    //public class NodeToExpressionVisitor : QueryNodeVisitor<Expression>
    //{

    //    public override Expression Visit(SingleValuePropertyAccessNode nodeIn)
    //    {
    //        return Expression.Constant(null);
    //    }

    //    public override Expression Visit(ConvertNode nodeIn)
    //    {
    //        return nodeIn.Accept(this);
    //    }

    //    public override Expression Visit(SingleValueFunctionCallNode nodeIn)
    //    {
    //        return nodeIn.Arguments.OfType<SingleValueFunctionCallNode>().First().Accept(this);
    //    }
    //}
}
