﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Threading.Tasks;

namespace test_mvc_webapi_ef_ng.Models
{
    public class Database : IDisposable
    {
        private db.Context _db = null;// new db.Context();
        private InMemDb _list = null;

        public Database(InMemDb db)
        {
            _list = db;
        }

        public DbResult<db.SystemInstance> QueryAll()
        {
            //var query = from s in _db.Systems orderby s.Id select s;
            var query = from s in _list._db orderby s.Id select s;

            return new DbResult<db.SystemInstance> { QueryResult = query.AsQueryable(), Count = query.Count(), Result = query.ToList() };
        }

        public DbResult<db.SomeComplex> QuerySome()
        {
            var list = new List<db.SomeComplex> { new db.SomeComplex { Id = 1, Name = "scpl" }, new db.SomeComplex { Id = 2, Name = "scpl 2" }, new db.SomeComplex { Id = 3, Name = "scpl 3" } };

            return new DbResult<db.SomeComplex> { QueryResult = list.AsQueryable(), Count = list.Count(), Result = list };
        }

        public DbResult<db.SystemInstance> Select(string sortColumn, bool descending, int skip, int take)
        {
            var q = from s in _db.Systems orderby s.Id select s;

            var provider = OrderExpressionFactory.Create<db.SystemInstance>(sortColumn);
            var query = provider.GetOrderByExpression<db.SystemInstance>(q, descending);

            return new DbResult<db.SystemInstance> { Result = query.Skip(skip).Take(take), Count = q.Count() };
        }

        public DbResult<db.SystemInstance> Select(int id)
        {
            var q = from s in _db.Systems where s.Id == id select s;

            return new DbResult<db.SystemInstance> { Result = q, Count = q.Count() };
        }

        public bool Insert(db.SystemInstance system)
        {
            bool result = false;

            if (system != null)
            {
                system.Id = _list._db.Count + 1;
                _list._db.Add(system);
                //_db.Systems.Add(system);
                //_db.SaveChanges();

                result = true;
            }

            return result;
        }

        public bool Update(db.SystemInstance system)
        {
            bool result = false;

            if (system != null)
            {
                var s = _list._db.SingleOrDefault(e => e.Id == system.Id);
                if (s != null)
                {
                    s.Name = system.Name;
                    //_db.Entry(system).State = System.Data.Entity.EntityState.Modified;
                    //_db.SaveChanges();

                    result = true;
                }
            }

            return result;
        }

        public bool Delete(int id)
        {
            bool result = false;

            if (id > 0)
            {
                var system = _list._db.FirstOrDefault(e => e.Id == id);
                //var system = _db.Systems.Find(id);
                if (system != null)
                {
                    _list._db.Remove(system);
                    //_db.Entry(system).State = System.Data.Entity.EntityState.Deleted;
                    //_db.SaveChanges();

                    result = true;
                }
            }

            return result;
        }


        #region IDisposable Support
        private bool disposedValue = false; // To detect redundant calls

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    // TODO: dispose managed state (managed objects).
                    if (_db != null)
                    {
                        _db.Dispose();
                    }
                }

                // TODO: free unmanaged resources (unmanaged objects) and override a finalizer below.
                // TODO: set large fields to null.

                disposedValue = true;
            }
        }

        // TODO: override a finalizer only if Dispose(bool disposing) above has code to free unmanaged resources.
        // ~Database() {
        //   // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
        //   Dispose(false);
        // }

        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
            Dispose(true);
            // TODO: uncomment the following line if the finalizer is overridden above.
            // GC.SuppressFinalize(this);
        }
        #endregion
    }


    public class DbResult<T>
    {
        public IQueryable<T> QueryResult { get; set; }
        public IEnumerable<T> Result { get; set; }
        public int Count { get; set; }
    }



    public interface IOrderExpressionProvider
    {
        IOrderedQueryable<TEntity> GetOrderByExpression<TEntity>(IOrderedQueryable<TEntity> query, bool descending);
        string ColumnName { get; set; }
        bool IsInvalid { get; }
    }


    public static class OrderExpressionFactory
    {
        private static Dictionary<Type, IOrderExpressionProvider> _provider = new Dictionary<Type, IOrderExpressionProvider>();
        private static OrderExpressionProvider<string> _default = new OrderExpressionProvider<string>();
        private static IOrderExpressionProvider _invalid = OrderExpressionProvider<string>.InvalidInstance;

        static OrderExpressionFactory()
        {
            register();
        }

        public static IOrderExpressionProvider Create<TEntity>(string columnName)
        {
            Type type = typeof(TEntity);

            PropertyInfo property = type.GetProperty(columnName ?? string.Empty, BindingFlags.Public | BindingFlags.Instance | BindingFlags.DeclaredOnly | BindingFlags.IgnoreCase);

            if (property == null)
            {
                return _invalid;
            }
            else
            {
                Type columnType = property.PropertyType;

                var provider = _provider.ContainsKey(columnType) ? _provider[columnType] : _default;
                provider.ColumnName = property.Name;

                return provider;
            }
        }

        private static void register()
        {
            _provider[typeof(int)] = new OrderExpressionProvider<int>();
        }
    }


    public class OrderExpressionProvider<TColumn> : IOrderExpressionProvider
    {
        public string ColumnName { get; set; }

        public IOrderedQueryable<TEntity> GetOrderByExpression<TEntity>(IOrderedQueryable<TEntity> query, bool descending)
        {
            if (query == null)
            {
                return null;
            }
            else if (this.IsInvalid)
            {
                return query;
            }

            ParameterExpression parameter = Expression.Parameter(typeof(TEntity));
            Expression body = Expression.Property(parameter, ColumnName);

            var lambda = Expression.Lambda<Func<TEntity, TColumn>>(body, parameter);

            return descending ? query.OrderByDescending<TEntity, TColumn>(lambda) : query.OrderBy<TEntity, TColumn>(lambda);
        }

        public bool IsInvalid { get { return this.GetType().GetGenericArguments().Any(t => t.Equals(typeof(OrderExpressionProvider<string>.Invalid))); } }

        public static readonly IOrderExpressionProvider InvalidInstance = new OrderExpressionProvider<Invalid>();


        private interface Invalid
        {
        }
    }
}
